

class ScenarioBuilder:
    
    def __init__(self):
        pass

    def withGrid(self, x, y):
        return self
    
    def withRobot(self, name):
        return RobotBuilder(self)



class RobotBuilder:
    
    def __init__(self, parentBuilder):
        self.parentBuilder = parentBuilder
        pass
    
    def done(self):
        return self.parentBuilder

class GoalBuilder():
    def __init__(self, name):
        self.parentBuilder = parentBuilder
        self.refinementType = None
        self.refinements = []
        pass

    def done(self):
        return self.parentBuilder

    def refinedBy(refinementType, refinements):
        pass

    def withGoal(name):
        self.name = name
    
    







import pytest
import time



from gmrs.utils.model import constant_outcomes

from gmrs.planner.operators.sequential import Sequential, SEQ_OP
from gmrs.planner.submodel import MultipleConstantEvaluationsSubmodel
from gmrs.planner.submodel import ConstantEvaluationSubmodel

from gmrs.evaluation.evaluation import Evaluation

from gmrs.objective.probability import Probability

from gmrs.planner.estimator import Estimator, estimatorFactory
from .utils import evaluationIn


estimator: Estimator = estimatorFactory(
    operators=[Sequential],
    objective_systems=[Probability])


def gen_mult_constant_eval_sub_model(num_tasks, num_confs_per_task):
    submodels = []
    for task in range(0, num_tasks):
        evals = []
        for conf in range(0, num_confs_per_task):
            eval_ = Evaluation({f'conf_t{task}': conf},
                               constant_outcomes(0.99))
            evals.append(eval_)
        m = MultipleConstantEvaluationsSubmodel(*evals)
        submodels.append(m)
    return submodels


def test_tasks_with_multi_confs_submodels():
    num_tasks = 10
    num_confs_per_task  = 4
    space = num_confs_per_task ** num_tasks
    print(f'space of {space}')
    seq_five_of_ten_confs_submodels = gen_mult_constant_eval_sub_model(
        num_tasks=num_tasks, num_confs_per_task=num_confs_per_task)

    start = time.time()
    evaluations = estimator.eval_op(SEQ_OP, seq_five_of_ten_confs_submodels)
    setup = time.time()
    num_evaluations = 0
    times = [None for _ in range(0, 10)]
    for evaluation in evaluations:
        sr = Probability.success_rate(evaluation.outcomes)
        # if num_evaluations > 1000:
        #     break
        
        if num_evaluations < 10:
            times[num_evaluations] = time.time()
        num_evaluations += 1
        # print(evaluation.config)
        # print(sr)
    end = time.time()
    print(f'{(setup - start):.6f} setup')
    print(f'{(times[0] - start):.6f} first evaluation d{times[0]-setup}')
    print(f'{(end - start):.6f} end time')

    last = setup
    for t in times:
        ns = (t - last)*1000*1000
        print(f'others {ns:.2f} ns')
        last = t

    assert num_evaluations == space




class SkillRepository:
    def __init__(self):
        self.repo = {}

    def queryFromTaskName(self, taskName):
        ref = self.normalizeName(taskName)
        return ref

    def register(self, taskName, skill):
        ref = self.normalizeName(taskName)
        if self.repo[ref] is None:
            self.repo[ref] = []
        self.repo[ref].append(skill)

    def normalizeName(self, taskName):
        return taskName




from gmrs.model.capability import Capability
from gmrs.planner.sill_base import SkillBase
from gmrs.model.robot import Robot
from gmrs.model.property import Property


class property:
    def distance(value, unit):
        return Property('distrance', value, unit)

    def speed(value, unit):
        return Property('speed', value, unit)


def capability(label, **properties):
    cap = Capability(label)
    cap.addProperties(**properties)
    return cap


def skill(id):
    return SkillBase(id)


def robot(id):
    return Robot(id)



from gmrs.graph.node import Node


class Context(Node):
    def addPropertyWithClassAndLabel(self, _class, label, value):
        if self[_class] is None:
            self[_class] = {}

        self[_class][label] = value

    def get(self, _class, label):
        if self[_class] is None:
            return None
        return self[_class][label]



from enum import Enum


class Result(Enum):
    FAILURE = 0
    SUCCESS = 1


class Outcome:
    # type = None
    def __init__(self, p=None, res: Result = None, **rewards):
        self.p = p
        self.res = res
        for key, value in rewards.items():
            setattr(self, key, value)


def is_success(outcome: Outcome):
    return outcome.res == Result.SUCCESS


def is_failure(outcome: Outcome):
    return outcome.res == Result.FAILURE



from gmrs.evaluation.outcome import Outcome
from gmrs.objective.objective_system import ObjectiveSystem


def seq_agg_all(res_out: Outcome,
                outcome1: Outcome, outcome2: Outcome,
                quality_operators: [ObjectiveSystem]):
    ''' aggregate properties of outcomes occurring sequentially '''
    for q in quality_operators:
        q.seq_agg(res_out, outcome1, outcome2)





import pytest


def evaluationIn(config, rate, *assepted_pairs):
    found = False
    for a in assepted_pairs:
        if a[0] == config:
            found = True
            assert a[1] == pytest.approx(rate)
            break
    if not found:
        raise Exception('evaluation with config not expected')



import math

from gmrs.evaluation.outcome import Result, Outcome, is_failure
from gmrs.evaluation.evaluation import Evaluation
from gmrs.planner.submodel import SubModel

from .operator import Operator
from .ievaluator import IEvaluator

SEQ_OP = 'SEQ'


class Sequential(Operator, IEvaluator):
    code = SEQ_OP

    def __init__(self, base):
        super().__init__(base)

    def next_node(self, node, outcome):
        # any task failure within a sequential compound lead to failure
        if is_failure(outcome):
            return Result.FAILURE

        index = self.nodes.find(node)
        if index < len(self.nodes):
            return self.node[index + 1]
        else:
            return Result.SUCCESS

    def merge_outcomes(self, outcomes1, outcomes2) -> Outcome:
        for o1 in outcomes1:
            if is_failure(o1):
                yield o1
            else:
                for o2 in outcomes2:
                    out = Outcome()
                    self.base.seq_agg(out, o1, o2)

                    out.res = (
                        Result.FAILURE
                        if is_failure(o2)
                        else Result.SUCCESS)

                    yield out

    def eval(self, submodels: SubModel) -> [Evaluation]:
        return self.seq_eval(submodels)

    def seq_eval(self, submodels: SubModel) -> [Evaluation]:
        '''
        Evaluate submodels and aggregate results
        '''

        if len(submodels) == 1:
            return self.base.eval_single(submodels[0])
        else:
            # divide
            end = len(submodels)
            mid = math.floor(end/2)
            eval_l = self.seq_eval(submodels[0:mid])
            eval_r = self.seq_eval(submodels[mid:end])
            # conquer
            agg = self.base.combineGen(eval_l, eval_r,
                                       self.aggregate_evaluations)
            # agg = self.aggregate_evaluations(eval_l, eval_r)
            return agg

    def aggregate_evaluations(self, evals1: [Evaluation], evals2: [Evaluation]):
        res_configs = self.base.merge_configs(evals1.config, evals2.config)
        res_outs = self.merge_outcomes(evals1.outcomes, evals2.outcomes)
        return Evaluation(config=res_configs, outcomes=res_outs)




"""Containers module."""

import logging.config
from dependency_injector import containers, providers

from ..estimator import Estimator
from ..operators.sequential import Sequential


class Container(containers.DeclarativeContainer):

    config = providers.Configuration()

    configure_logging = providers.Callable(
        logging.config.fileConfig,
        fname='logging.ini',
    )

    sequential_operator = providers.Factory(
        Sequential
    )

    estimator_operator = providers.Factory(
        Estimator
    )





import abc

from gmrs.evaluation.evaluation import Evaluation
from ..submodel import SubModel


class IEvaluator:
    """
    docstring
    """

    @abc.abstractclassmethod
    def eval(self, submodels: [SubModel]) -> [Evaluation]:
        pass



from typing import Type
from collections.abc import Iterable
import itertools

from gmrs.evaluation.outcome import Outcome

from gmrs.planner.submodel import SubModel
from gmrs.planner.submodel import ConstantEvaluationSubmodel, RefinedSubmodel
from gmrs.planner.submodel import MultipleConstantEvaluationsSubmodel

from gmrs.objective.objective_system import ObjectiveSystem
from .operators.ievaluator import IEvaluator


class Estimator:
    objective_systems: [ObjectiveSystem] = []
    operators = dict()

    def __init__(self):
        pass

    def eval_op(self, op, submodels: [SubModel]):
        operator: IEvaluator = self.operators[op]
        return operator.eval(submodels)

    def eval_single(self, model):
        if isinstance(model, ConstantEvaluationSubmodel):
            return model.evaluation
        elif isinstance(model,  MultipleConstantEvaluationsSubmodel):
            return model.eval()
        elif isinstance(model, RefinedSubmodel):
            op = model.op
            return self.eval_op(op, model.refinements)
        else:
            mt = model.__class__.__name__
            err = f'trying to evaluate not supported model type {mt}'
            raise Exception(err)

    def combineGen(self, gena, genb, func):
        a = gena if isinstance(gena, Iterable) else [gena]
        b = genb if isinstance(genb, Iterable) else [genb]
        return itertools.starmap(func, itertools.product(a, b))

    def seq_agg(self, res_out: Outcome,
                outcome1: Outcome, outcome2: Outcome):
        ''' aggregate properties of outcomes occurring sequentially '''
        for q in self.objective_systems:
            q.seq_agg(res_out, outcome1, outcome2)

    def merge_configs(self, config1: dict, config2: dict) -> dict:
        nconfig = dict()
        nconfig.update(config1)
        for key, value in config2.items():
            c1_value = nconfig.get(key)
            if c1_value is not None and c1_value != value:
                # TODO handle warn
                print('warn: merging configs with conflicting name space: {key}')
        nconfig.update(config2)
        return nconfig


# Factory

def estimatorFactory(
            operators: [],  # operators: [Type[Operator]],
            objective_systems: [Type[ObjectiveSystem]]) -> Estimator:

    estimator = Estimator()

    for operator in operators:
        op = operator(estimator)
        estimator.operators[op.code] = op

    estimator.objective_systems = objective_systems
    return estimator



#!/usr/bin/python
from gmrs.model.mission.operator import Operator


def composit_task(ref_type, *tasks):
    op = Operator(ref_type)
    for t in tasks:
        op.addEdge(t)


def add_refinement(origin, refinement):
    op = Operator(refinement[0])
    op.addEdges(refinement[1:])


def set_location(node, location):
    pass




class MapOfDistance():
    _map = {}

    def add_direct_segment(self, locationA, locationB, distance):
        if self._map[locationA.name] is None:
            self._map[locationA.name] = {}
        self._map[locationA.name][locationB.name] = distance




from .submodel import SubModel


def get_constant_fnc(p, t=1):
    assert p < 1 and p > 0
    return [(p, t), (1-p, t)]


def evaluations_iterator(subModel: SubModel):
    configs = subModel.get_configs()
    for config in configs:
        yield (config, subModel.eval(config))


def submodels_evaluations_combinations(*models):
    header_iter = evaluations_iterator(models[0])
    if len(models) == 1:
        # single configuration
        for model_eval in header_iter:
            yield model_eval
    else:
        # combined configurations
        for header_eval in header_iter:
            rest_iter = submodels_evaluations_combinations(*models[1:])
            for rest_eval in rest_iter:
                if isinstance(rest_eval, tuple):
                    yield (header_eval, *rest_eval)
                else:
                    yield (header_eval, rest_eval)



from gmrs.evaluation.outcome import Result


class VariabilityModelBuilder():
    macro_lib = None
    skill_lib = None

    def instantiate_variability_model(self, task, mission, unit,
                                      environment_context):
        context = environment_context.extend(task.parameters, unit.context,
                                             environment_context)
        # phase one
        controller = controller()
        gvm = self.bind_node(mission, context, controller)
        write_graph(gvm)

        # phase two
        def should_stop(model=model, controller=controller):
            return controller.checked_all()

        while not should_stop():
            controller.expand()
            curr_plan = controller.last_plan()
            print(evaluation(curr_plan))

        plan = controller.last_plan()

    def bind_node(self, node, context, controller):
        if is_leaf(node):
            if self.is_macro(node):
                macro = self.resolve_macro(node)
                return self.bind_node(macro, context, macro_lib, skill_lib)
            elif is_skill(node):
                # TODO build alternative if multiple skills
                return self.bind_skills(node)
        else:
            refinement = self.get_refinement(node)
            operator = self.get_operator(refinement)
            children = self.get_children(refinement)
            bound_self.bind_children(operator, children,
                                                  controller):
                

            
    def bind_children(self, operator, children, context, controller):
        def has_next():
            return True
        
        first_child = operator.get_first(children)
        queue = (first_child, context)

        for bind in queue:
            child_node, context = bind

            # sequential context
            bound_child_node = self.bind_node(child_node, context)
            result = bound_child_node.check(context)
            if not result.is_achievable:
                print('State not Handled - not achievable task. TODO')
                continue
            
            possible_outcomes = result.possible_outcomes
                
            for po in possible_outcomes:
                next_node = operator.next_node(child_node, po)
                next_context = context.extend(po)
                
                if is_end(next_node):


                new_binding = (next_node, next_context)
                queue.append(new_bindings)
        
        return next_context


    def is_leaf(node):
        return False


    def bind_child(self, child):
        return None

    def resolve(self, refinements):
        return None

    def parseName(self, nodeName):
        return None

    def resolve_macro(self, node):
        pass

    def is_skill(self, node):
        pass

    def bind_skills(self, node):
        pass

    def init_context_from_scenario(scenario):
        pass

    def write_graph(self):
        if self.is_write_graph_enable:
            print('writting graph')



from gmrs.model.property import Property


class WorldMap:
    def __init__(self, args):
        for arg in args:
            map['poi1']['poi2'] = Property('distrance', args['distance'][0], args['distance'][1])

    def getDistance(self, poi1, poi2):
        return self[poi1][poi2]


def create_map(args):
    return WorldMap(args)




import pytest

from gmrs.planner.submodel import ConstantEvaluationSubmodel, RefinedSubmodel
from gmrs.utils.model import constant_outcomes

from gmrs.planner.operators.sequential import Sequential, SEQ_OP
from gmrs.evaluation.evaluation import Evaluation

from gmrs.objective.probability import Probability

from gmrs.planner.estimator import Estimator, estimatorFactory


# container = containers.Container()
# container.configure_logging()
# container.config.from_ini('config.ini')


estimator: Estimator = estimatorFactory(
            operators=[Sequential],
            objective_systems=[Probability])

evaluationA = Evaluation({'config1': 'a'}, constant_outcomes(0.9))
evaluationB = Evaluation({'config2': 'b'}, constant_outcomes(0.8))
evaluationC = Evaluation({'config3': 'c'}, constant_outcomes(0.7))

two_submodels = [
    ConstantEvaluationSubmodel(evaluationA),
    ConstantEvaluationSubmodel(evaluationB)
    ]

three_submodels = [
    ConstantEvaluationSubmodel(evaluationA),
    ConstantEvaluationSubmodel(evaluationB),
    ConstantEvaluationSubmodel(evaluationC)
    ]

submodels_with_mixed_refinement = [
    ConstantEvaluationSubmodel(evaluationA),
    ConstantEvaluationSubmodel(evaluationB),
    RefinedSubmodel(SEQ_OP,
                    [ConstantEvaluationSubmodel(evaluationA),
                     ConstantEvaluationSubmodel(evaluationB)])
]


def test_eval_seq_two_submodels():
    evaluation = next(estimator.eval_op(SEQ_OP, two_submodels))
    sr = Probability.success_rate(evaluation.outcomes)
    assert 0.72 == pytest.approx(sr)
    print(sr)


def test_eval_seq_three_submodels():
    evaluation = next(estimator.eval_op(SEQ_OP, three_submodels))
    sr = Probability.success_rate(evaluation.outcomes)
    assert 0.504 == pytest.approx(sr)
    print(sr)


def test_eval_seq_with_mixed_refinement():
    evaluation = next(estimator.eval_op(SEQ_OP, submodels_with_mixed_refinement))
    sr = Probability.success_rate(evaluation.outcomes)
    assert 0.5184 == pytest.approx(sr)
    print(sr)


# from gmrs.planner.quality.time import Time



from gmrs.evaluation.outcome import Outcome


class Evaluation():
    config: dict = None
    outcomes = []

    def __init__(self, config: dict, outcomes: [Outcome]):
        self.config = config
        self.outcomes = outcomes




from gmrs.graph.node import Node


class Task(Node):
    def __init__(self, id, fnc=lambda: []):
        super().__init__(self)
        self.id = id
        self.fnc = fnc


class LeafTask(Task):
    def __init__(self, label):
        self.skill = label
        pass


def extract_vars(label):
    m = re.search(r'\[(.*?)\]', label)
    return m.group(0)


def refinement():
    pass



from gmrs.evaluation.outcome import Outcome
from gmrs.services.base import BaseService


class ObjectiveSystem:
    @staticmethod
    def seq_agg(res: Outcome, outcome1: Outcome, outcome2: Outcome):
        pass


class BaseObjectiveSystem(ObjectiveSystem, BaseService):
    pass



from gmrs.evaluation.outcome import Outcome, is_success
from .objective_system import ObjectiveSystem


class Probability(ObjectiveSystem):

    @staticmethod
    def seq_agg(res_outcome: Outcome, outcome1: Outcome, outcome2: Outcome):
        res_outcome.p = outcome1.p * outcome2.p

    @staticmethod
    def success_rate(outcomes: [Outcome]):
        res = 0.0
        for out in outcomes:
            if is_success(out):
                res += out.p
        return res



import itertools


class each_reset_wrapper:
    def __init__(self, gen):
        self.gen = gen


class iterFib:
    called = 0

    def fibo(self, x):
        print('init')
        self.called += 1
        i, j = 1, 1
        yield j
        while j <= x:
            _next = i + j
            yield _next
            print('x')
            self.called += 1
            i = j
            j = _next
        print('end')


def test_gen_conbination():
    fib1 = iterFib()
    fib2 = iterFib()

    def sum(seqa, seqb):
        return itertools.starmap(lambda x, y: x+y, itertools.product(seqa, seqb))
            
    # for i, j in itertools.product(fib1.fibo(3), fib2.fibo(3)):
    #     print(i)
    #     print(j)
    print('lets')
    for x in sum(fib1.fibo(3), fib2.fibo(3)):
        print(x)
    
    print('----')
    print(fib1.called)
    print(fib2.called)




import abc
from enum import Enum

from gmrs.graph.node import Node


class SkillStatus(Enum):
    unbound = 0
    check_and_bind = 1
    expand = 2
    wait = 3
    cancel = 4
    deploy = 5
    execute = 6
    report = 7
    end = 8


class SkillBase:
    name = '_'
    status = SkillStatus.unbound

    def __init__(self, *params):
        self.init(*params)

    @abc.abstractmethod
    def check_and_bind(self):
        pass

    @abc.abstractmethod
    def expand(self):
        pass

    @abc.abstractmethod
    def wait(self):
        pass

    @abc.abstractmethod
    def cancel(self):
        pass

    @abc.abstractmethod
    def deploy(self):
        pass

    @abc.abstractmethod
    def execute(self):
        pass

    @abc.abstractmethod
    def report(self):
        pass

    @abc.abstractmethod
    def end(self):
        pass

    def to_node(self):
        return Node(self.name)



class Node:
    def __init__(self, label=None, type=None, leaf=False):
        self.label = label
        self.type = type or self.__class__.__name__
        self.leaf = leaf
        self.edges = []
        self.typedEdges = {}

    def addEdge(self, edge):
        self.edges.append(edge)

    def addEdges(self, edges):
        for edge in edges:
            self.addEdge(edge)

    def addTypedEdge(self, type, edge):
        self.addEdge(edge)
        self.typedEdges[type] = edge

    def addProperties(self, **properties):
        if not hasattr(self, 'properties'):
            self.properties = {}
        self.properties.update(properties)


class TreeNode(Node):
    def __init__(self):
        super().__init__(self)
        self.parent = None

    def add_parent(self, node, edge_attrs):
        self.parent = Edge(self, node, edge_attrs)


class Edge:
    def __init__(self, nodeA, nodeB, attrs=None):
        self.nodeA = nodeA
        self.nodeB = nodeB
        self.attrs = attrs



import pytest

from gmrs.model.mission import Mission
from gmrs.model.goal import Goal
from gmrs.model.task import Task

from gmrs.model.location import Location

from gmrs.model.robot import Robot

from gmrs.model.capability import Capability

from gmrs.operators.operators import SEQ
from gmrs.utils.task_refinement_utils import rft
from gmrs.environment.map_of_distance import MapOfDistance


from ..minimal_world.skill_visit import VisitWithMoveBase


# world
# locations
class world:
    robots_room = Location('robots_room')
    room = Location('room')
    map = MapOfDistance()

    map.add_direct_segment(robots_room, room, 100)


class mission:

    # Task Refinement
    # tidy
    identify_out_of_place_objects = Task('Identify out of place objects')
    move_objects = Task('Move Objects')
    tidy_objects = rft(Task('Tidy Objects of [room]'),
                       SEQ, identify_out_of_place_objects, move_objects)

    # clean
    sweep_floor = Task('Sweep Floor')
    uv_light_sterilize = Task('UV-Light Sterilize')

    clean_room = rft(Task('Clean [room]'),
                     SEQ, sweep_floor, uv_light_sterilize)

    # root
    clean_and_tidy_room = rft(Task('Clean and Tidy [room]'),
                              SEQ, tidy_objects, clean_room)

    goal = Goal(clean_and_tidy_room)
    objective = None


class robots:
    move_configurations = {
        'slow': {'speed': 10},
        'default': 'slow',
    }

    move = Capability('move', configs=move_configurations)

    skill = VisitWithMoveBase()

    r1 = Robot('r1',
               location=world.robots_room,
               capabilities=[move],
               skills=None)


class scenario:
    world = world
    mission = mission
    robots = robots


@pytest.fixture
def clean_room_scenario():
    return scenario




import abc

from gmrs.evaluation.evaluation import Evaluation

from ..submodel import SubModel
from ..estimator import Estimator


class Operator:
    base: Estimator = None

    def __init__(self, base):
        self.base = base

    @abc.abstractclassmethod
    def eval(self, submodel: SubModel) -> [Evaluation]:
        pass




import unittest
from enum import Enum

from gmrs.model.context import Context
from gmrs.model.task import LeafTask
from gmrs.model.robot import Robot
from gmrs.model.factory import capability, property
from gmrs.model.world_map import create_map

from gmrs.planner.skill_base import SkillBase
from gmrs.evaluate import evaluate
from gmrs.evaluate import get_skill
from gmrs.utils.physics import estimate_dt


from .move_capability import MoveCapability


class VisitWithMoveBase(SkillBase):

    provides = 'Visit [destination]'

    pre_conditions = {
        'has_capabilities': [MoveCapability]
    }

    parameters = {
        'destination': 'location',
    }

    configurations = {}
    origin, detination, world_map, paths_generator = None

    def __init__(self, logger, robot, task_parameters, world_model, context):
        # safe init
        self.name = self.format_name(provides)
        self.logger = logger
        # check achievability, raise error if something wrong
        #self.init(robot, task_parameters, world_model, context)

    def bind(self, robot, task_parameters, world_model, context):
        ''' 
        bind from robot (state, capabilities), task parameters, world map 
        '''
        origin = self._get_required('location', context,
                                    default=robot.initial_location)
        detination = self._get_required('destination', task_parameters)
        world_map = self._get_required('map', self.world_model)
        
        paths_generator = self.world_model.map.get_paths(origin, detination)

        # robot capabilities
        move = get_capability(robot, MoveCapability)

        # configuration options
        init_configurations(paths=paths_generator,
                            move_configurations=move.configs)
    
    def get_post_condition(self):
        # post condition
        post_conditions = {
            'location': detination
        }
        return post_conditions
    
    def init_configurations(self, paths, move_configurations):
        pass

    def config_iterator():
        
        yield config

    def get_heuristic(self):
        return 0

    def expand(self):
        for config in self.config_iterator(self.configurations):
            yield self.eval(*config)

    def _get_required(self, name, container, default):
        value = container[name] if container[name] is not None else default
        if value is None:
            fail_code = f'get-required-{name}'
            self.logger(fail_code)
            raise Exception(fail_code)

    def get_capability(self, robot, capability):
        pass

    # function on the injected dependencies
    def eval(max_speed_options, current_location, place, map):
        # base calculation
        time = estimate_dt(
            distance=map.getDistance(current_location, place),
            speed=max_speed)

        battery = map.getDistance(current_location, place) * 
        # possible outcomes - probabily p should also be a function of time
        # staticaly assume that every robot is capable of 
        # 3 speeds to improve the readability
        return {
            'speed=["slow"]': [
                    {'p': 0.99, status: SUCC, time: time},
                    {'p': 0.01, status: ERR, time: time}
            ],
            'speed=["medium"]': [
                {'p': 0.99, status: SUCC, time: time},
                {'p': 0.01, status: ERR, time: time}
            ],
            'speed=["fast"]': [
                {'p': 0.99, status: SUCC, time: time},
                {'p': 0.01, status: ERR, time: time}
            ],
            }
        return (vars, eval)

    def task_end(self):
        pass



required_capabilities = ['self_localization', 'move']

skill_visit = Skill('Visit [place]',
                    assumme=[required_capabilities],
                    guarantee=['location:[place]'],
                    nodes=['visit_skill']
                    fnc=visit_fnc())


# task decomposition model with a single task
task_decomposition = LeafTask('Visit [place]')
#MPCTL objective
#constraints = 

# Runtime ####
request = 'Visit'



import abc


class SubModel:

    @abc.abstractmethod
    def eval(self):
        '''
        return ConfigEvaluation
        '''
        pass


class ConstantEvaluationSubmodel(SubModel):
    def __init__(self, evaluation):
        self.evaluation = evaluation


class MultipleConstantEvaluationsSubmodel(SubModel):
    def __init__(self, *evaluations):
        self.evaluations = evaluations

    def eval(self):
        return self.evaluations


class RefinedSubmodel(SubModel):
    def __init__(self, op, refinements: [SubModel]):
        self.op = op
        self.refinements = refinements





coins = [1, 5, 10, 25, 50, 100, 200, 500, 1000, 2000, 5000, 10000]

ccount = [i for i in range(10000 + 1)]
cselection = [None for i in range(10000 + 1)]

for c in coins:
    ccount[c] = 1
    cselection[c] = (c)


for i in range(10000):
    if ccount[i] <= 1:
        continue
    else:
        t = ccount[i]
        for j in range(0, i):
            s = ccount[j] + ccount[i-j]
            if s < t:
                t = s
                cselection[i] = (cselection[j], cselection[i-j])
        ccount[i] = t
        print(f'{i}:{t}')


max = 0
maxi = 0
for i in range(10000):
    if(ccount[i] > max):
        max = ccount[i]
        maxi = i

print(f'>{maxi}:{max}')
print(f'>{maxi}:{cselection[maxi]}')



from gmrs.planner.skill_base import SkillBase
from gmrs.objective.objective_system import ObjectiveSystem


def evaluate(model, context):
    outcomes = None
    if is_leaf(model):
        skill = get_skill(model, context)
        outcomes = evaluate_task(model, skill)
    else:
        pass

    outcomes = evaluate(model, context)
    return outcomes


def evaluate_task(task, skill):
    pass


def get_task(model):
    return model


def is_leaf(model):
    return True


def get_skill(task, context):
    return context.get(Skill, task['label'])




from gmrs.evaluation.outcome import Outcome
from .objective_system import ObjectiveSystem


class Time(ObjectiveSystem):

    id = 't'

    @staticmethod
    def seq_agg(res_outcome: Outcome, outcome1: Outcome, outcome2: Outcome):
        res_outcome[id] = outcome1[id] + outcome2[id]



import pytest

from gmrs.utils.model import constant_outcomes

from gmrs.planner.operators.sequential import Sequential, SEQ_OP
from gmrs.planner.submodel import MultipleConstantEvaluationsSubmodel
from gmrs.planner.submodel import ConstantEvaluationSubmodel

from gmrs.evaluation.evaluation import Evaluation

from gmrs.objective.probability import Probability

from gmrs.planner.estimator import Estimator, estimatorFactory
from .utils import evaluationIn


estimator: Estimator = estimatorFactory(
    operators=[Sequential],
    objective_systems=[Probability])

moveSlow = Evaluation({'speed': 'slow'}, constant_outcomes(0.9, time=3))
moveFast = Evaluation({'speed': 'fast'}, constant_outcomes(0.8, time=2))
actionX = Evaluation({'config_action': 'x'}, constant_outcomes(0.7, time=8))
actionY = Evaluation({'config_action': 'y'}, constant_outcomes(0.6, time=9))

twin_then_single_submodels = [
    MultipleConstantEvaluationsSubmodel(moveSlow, moveFast),
    ConstantEvaluationSubmodel(actionX)
]

twin_then_twin_submodels = [
    MultipleConstantEvaluationsSubmodel(moveSlow, moveFast),
    MultipleConstantEvaluationsSubmodel(actionX, actionY),
]


def test_seq_twin_with_single():
    evaluations = estimator.eval_op(SEQ_OP, twin_then_single_submodels)
    for evaluation in evaluations:
        sr = Probability.success_rate(evaluation.outcomes)
        evaluationIn(evaluation.config, sr,
                     ({'speed': 'slow', 'config_action': 'x'}, 0.63),
                     ({'speed': 'fast', 'config_action': 'x'}, 0.56))
        print(sr)


def test_twin_with_twin():
    evaluations = estimator.eval_op(SEQ_OP, twin_then_twin_submodels)
    for evaluation in evaluations:
        sr = Probability.success_rate(evaluation.outcomes)
        evaluationIn(evaluation.config, sr,
                     ({'speed': 'slow', 'config_action': 'x'}, 0.63),
                     ({'speed': 'fast', 'config_action': 'x'}, 0.56),
                     ({'speed': 'slow', 'config_action': 'y'}, 0.54),
                     ({'speed': 'fast', 'config_action': 'y'}, 0.48))
        print(sr)
